@@ -30,7 +30,9 @@
 /* Abbreviation length for HTML outputs */
 #define VI_HTML_ABBR_LEN 100
 /* Version as a string */
-#define VI_VERSION_STR "0.4a"
+#define VI_DATE_MAX 64
+/* Max length of a log entry date */
+#define VI_VERSION_STR "0.5"
 
 /*------------------------------- data structures ----------------------------*/
 
@@ -43,6 +45,7 @@ struct vih {
 	int hour[24];
 	int weekday[7];
 	int weekdayhour[7][24]; /* hour and weekday combined data */
+	int monthday[12][31]; /* month and day combined data */
 	struct hashtable visitors;
 	struct hashtable googlevisitors;
 	struct hashtable pages;
@@ -54,14 +57,18 @@ struct vih {
 	struct hashtable referersage;
 	struct hashtable date;
 	struct hashtable googledate;
+	struct hashtable month;
+	struct hashtable googlemonth;
 	struct hashtable agents;
 	struct hashtable googled;
 	struct hashtable googlevisits;
 	struct hashtable googlekeyphrases;
+	struct hashtable googlekeyphrasesage;
 	struct hashtable trails;
 	struct hashtable tld;
 	struct hashtable os;
 	struct hashtable browsers;
+	struct hashtable robots;
 	char *error;
 };
 
@@ -115,20 +122,29 @@ int Config_max_error404 = 20;
 int Config_max_agents = 20;
 int Config_max_googled = 20;
 int Config_max_google_keyphrases = 20;
+int Config_max_google_keyphrases_age = 20;
 int Config_max_trails = 20;
 int Config_max_tld = 20;
+int Config_max_robots = 20;
 int Config_process_agents = 0;
 int Config_process_google = 0;
-int Config_process_google_keyphrases;
+int Config_process_google_keyphrases = 0;
+int Config_process_google_keyphrases_age = 0;
 int Config_process_web_trails = 0;
 int Config_process_weekdayhour_map = 0;
+int Config_process_monthday_map = 0;
 int Config_process_referers_age = 0;
 int Config_process_tld = 0;
 int Config_process_os = 0;
 int Config_process_browsers = 0;
 int Config_process_error404 = 0;
 int Config_process_pageviews = 0;
+int Config_process_monthly_visitors = 1;
+int Config_process_robots = 0;
 int Config_graphviz_mode = 0;
+int Config_graphviz_ignorenode_google = 0;
+int Config_graphviz_ignorenode_external = 0;
+int Config_graphviz_ignorenode_noreferer = 0;
 int Config_tail_mode = 0;
 int Config_stream_mode = 0;
 int Config_update_every = 60*10; /* update every 10 minutes for default. */
@@ -143,6 +159,9 @@ struct vistring Config_prefix[VI_PREFIXES_MAX];
 
 /*----------------------------------- Tables ---------------------------------*/
 static char *vi_wdname[7] = {"Mo", "Tu", "We", "Th", "Fr", "Sa", "Su"};
+#if 0
+static int vi_monthdays[12] = {31, 29, 31, 30, 31, 30 , 31, 31, 30, 31, 30, 31};
+#endif
 
 /* -------------------------------- prototypes ------------------------------ */
 void vi_clear_error(struct vih *vih);
@@ -191,13 +210,17 @@ int vi_is_image(char *s)
 	    !memcmp(end-4, ".gif", 4) ||
 	    !memcmp(end-4, ".png", 4) ||
 	    !memcmp(end-4, ".ico", 4) ||
+	    !memcmp(end-4, ".swf", 4) ||
+	    !memcmp(end-3, ".js", 3) ||
 	    !memcmp(end-5, ".jpeg", 5) ||
 	    !memcmp(end-4, ".CSS", 4) ||
 	    !memcmp(end-4, ".JPG", 4) ||
 	    !memcmp(end-4, ".GIF", 4) ||
 	    !memcmp(end-4, ".PNG", 4) ||
 	    !memcmp(end-4, ".ICO", 4) ||
-	    !memcmp(end-4, ".JPEG", 5)) return 1;
+	    !memcmp(end-4, ".SWF", 4) ||
+	    !memcmp(end-3, ".JS", 3) ||
+	    !memcmp(end-5, ".JPEG", 5)) return 1;
 	return 0;
 }
 
@@ -356,6 +379,19 @@ int vi_is_weekend(char *s)
 	return 0;
 }
 
+#if 0
+/* Returns true if 'year' is a leap year. */
+int isleap(int year)
+{
+	int conda, condb, condc;
+
+	conda = (year%4) == 0;
+	condb = (year%100) == 0;
+	condc = (year%400) == 0;
+	return conda && !(condb && !condc);
+}
+#endif
+
 /* URL decoding and white spaces trimming function.
  * Input: the encoded string 's'.
  * Output: the decoded string written at 'd' that has room for at least 'n'
@@ -423,7 +459,7 @@ void vi_urlencode(char *d, char *s, int n)
 			n -= 6;
 		} else {
 			unsigned int t;
-			char *hexset = "0123456789abcdf";
+			char *hexset = "0123456789abcdef";
 
 			if (n < 3) break;
 			t = (unsigned) c;
@@ -512,7 +548,7 @@ void vi_ht_init(struct hashtable *ht)
 }
 
 /* Reset the weekday/hour info in the visitors handler. */
-void vi_reset_weekday_hour(struct vih *vih)
+void vi_reset_combined_maps(struct vih *vih)
 {
 	int i, j;
 
@@ -522,6 +558,9 @@ void vi_reset_weekday_hour(struct vih *vih)
 			vih->weekdayhour[j][i] = 0;
 	}
 	for (i = 0; i < 7; i++) vih->weekday[i] = 0;
+	for (i = 0; i < 31; i++)
+		for (j = 0; j < 12; j++)
+			vih->monthday[j][i] = 0;
 }
 
 /* Reset the hashtables from the handler, that are left
@@ -540,6 +579,7 @@ void vi_reset_hashtables(struct vih *vih)
 	ht_destroy(&vih->agents);
 	ht_destroy(&vih->googled);
 	ht_destroy(&vih->googlekeyphrases);
+	ht_destroy(&vih->googlekeyphrasesage);
 	ht_destroy(&vih->googlevisits);
 	ht_destroy(&vih->trails);
 	ht_destroy(&vih->tld);
@@ -547,13 +587,16 @@ void vi_reset_hashtables(struct vih *vih)
 	ht_destroy(&vih->browsers);
 	ht_destroy(&vih->date);
 	ht_destroy(&vih->googledate);
+	ht_destroy(&vih->month);
+	ht_destroy(&vih->googlemonth);
+	ht_destroy(&vih->robots);
 }
 
 /* Reset handler informations to support --reset option in
  * stream mode. */
 void vi_reset(struct vih *vih)
 {
-	vi_reset_weekday_hour(vih);
+	vi_reset_combined_maps(vih);
 	vi_reset_hashtables(vih);
 }
 
@@ -571,7 +614,7 @@ struct vih *vi_new(void)
 	vih->startt = vih->endt = time(NULL);
 	vih->processed = 0;
 	vih->invalid = 0;
-	vi_reset_weekday_hour(vih);
+	vi_reset_combined_maps(vih);
 	vih->error = NULL;
 	vi_ht_init(&vih->visitors);
 	vi_ht_init(&vih->googlevisitors);
@@ -586,12 +629,16 @@ struct vih *vi_new(void)
 	vi_ht_init(&vih->googled);
 	vi_ht_init(&vih->googlevisits);
 	vi_ht_init(&vih->googlekeyphrases);
+	vi_ht_init(&vih->googlekeyphrasesage);
 	vi_ht_init(&vih->trails);
 	vi_ht_init(&vih->tld);
 	vi_ht_init(&vih->os);
 	vi_ht_init(&vih->browsers);
 	vi_ht_init(&vih->date);
+	vi_ht_init(&vih->month);
 	vi_ht_init(&vih->googledate);
+	vi_ht_init(&vih->googlemonth);
+	vi_ht_init(&vih->robots);
 	return vih;
 }
 
@@ -885,17 +932,34 @@ void vi_process_date_and_hour(struct vih *vih, int weekday, int hour)
 	vih->weekdayhour[weekday][hour]++;
 }
 
+/* process the month and day information */
+void vi_process_month_and_day(struct vih *vih, int month, int day)
+{
+	if (month < 0 || month > 11 || day < 0 || day > 30) return;
+	vih->monthday[month][day]++;
+}
+
 /* Process unique visitors populating the relative hash table.
  * Return non-zero on out of memory. This is also used to populate
- * the hashtable used for the "pageviews per user" statistics. */
-int vi_process_visitors_per_day(struct vih *vih, char *host, char *agent, char *date, char *ref, char *req)
+ * the hashtable used for the "pageviews per user" statistics.
+ *
+ * Note that the last argument 'seen', is an integer passed by reference
+ * that is set to '1' if this is not a new visit (otherwise it's set to zero) */
+int vi_process_visitors_per_day(struct vih *vih, char *host, char *agent, char *date, char *ref, char *req, int *seen)
 {
-	char visday[VI_LINE_MAX], *p;
-	int res, host_len, agent_len, date_len;
+	char visday[VI_LINE_MAX], *p, *month = "fixme if I'm here!";
+        char buf[64];
+	int res, host_len, agent_len, date_len, hash_len;
+        unsigned long h;
 
+        /* Build an unique identifier for this visit
+         * adding together host, date and hash(user agent) */
 	host_len = strlen(host);
 	agent_len = strlen(agent);
 	date_len = strlen(date);
+        h = djb_hash(agent, agent_len);
+        sprintf(buf, "%lu", h);
+        hash_len = strlen(buf);
 	if (host_len+agent_len+date_len+4 > VI_LINE_MAX)
 		return 0;
 	p = visday;
@@ -903,9 +967,16 @@ int vi_process_visitors_per_day(struct vih *vih, char *host, char *agent, char *
 	*p++ = '|';
 	memcpy(p, date, date_len); p += date_len;
 	*p++ = '|';
-	memcpy(p, agent, agent_len); p += agent_len;
+	memcpy(p, buf, hash_len); p += hash_len;
 	*p = '\0';
 
+	if (Config_process_monthly_visitors) {
+		/* Skip the day number. */
+		month = strchr(date, '/');
+		if (!month) return 0; /* should never happen */
+		month++;
+	}
+
 	/* Visits with Google as referer are also stored in another hash
 	 * table. */
 	if (vi_is_google_link(ref)) {
@@ -914,6 +985,10 @@ int vi_process_visitors_per_day(struct vih *vih, char *host, char *agent, char *
 		if (res == 1) { /* new visit! */
 			res = vi_counter_incr(&vih->googledate, date);
 			if (res == 0) return 1; /* out of memory */
+			if (Config_process_monthly_visitors) {
+				res = vi_counter_incr(&vih->googlemonth, month);
+				if (res == 0) return 1; /* out of memory */
+			}
 		}
 	}
 	/* Populate the 'pageviews per visitor' hash table */
@@ -924,9 +999,17 @@ int vi_process_visitors_per_day(struct vih *vih, char *host, char *agent, char *
 	/* Mark the visit in the non-google-specific hashtable */
 	res = vi_counter_incr(&vih->visitors, visday);
 	if (res == 0) return 1; /* out of memory */
-	if (res > 1) return 0; /* visit alredy seen. */
+	if (res > 1) {
+		if (seen) *seen = 1;
+		return 0; /* visit alredy seen. */
+	}
+	if (seen) *seen = 0; /* new visitor */
 	res = vi_counter_incr(&vih->date, date);
 	if (res == 0) return 1;
+	if (Config_process_monthly_visitors) {
+		res = vi_counter_incr(&vih->month, month);
+		if (res == 0) return 1;
+	}
 	return 0;
 }
 
@@ -1055,7 +1138,6 @@ int vi_process_browsers(struct vih *vih, char *agent)
 		"w3m", "W3M",
 		"NATSU-MICAN", NULL,
 		"www.googlebot.com", "GoogleBot",
-		"www.google.com/bot", "GoogleBot",
 		"", "Unknown",
 		NULL, NULL,
 	};
@@ -1067,14 +1149,14 @@ int vi_process_browsers(struct vih *vih, char *agent)
 int vi_process_googled(struct vih *vih, char *req, char *agent, time_t age)
 {
 	if (!strstr(agent, "Googlebot") &&
-	    !strstr(agent, "googlebot")) return 0;
+            !strstr(agent, "googlebot")) return 0;
 	return vi_replace_if_newer(&vih->googled, req, age);
 }
 
 /* Process accesses with the referer from google.
  * This is used to populate the keyphrases hashtable.
  * TODO: url decoding */
-int vi_process_google_keyphrases(struct vih *vih, char *ref)
+int vi_process_google_keyphrases(struct vih *vih, char *ref, time_t age)
 {
 	char *s, *p, *e;
 	int res, page;
@@ -1104,27 +1186,52 @@ int vi_process_google_keyphrases(struct vih *vih, char *ref)
 	res = vi_counter_incr(&vih->googlekeyphrases, urldecoded);
 	if (e) *e = '&';
 	if (res == 0) return 1;
+	/* Process keyphrases by first time */
+	if (Config_process_google_keyphrases_age) {
+		if (vi_replace_if_older(&vih->googlekeyphrasesage,
+					urldecoded, age)) return 1;
+	}
 	return 0;
 }
 
+/* Process robots information. For visitors every client accessing
+ * to robots.txt is considered a robot.
+ * Returns 1 on out of memory, otherwise zero is returned. */
+int vi_process_robots(struct vih *vih, char *req, char *agent)
+{
+	if (strncmp(req, "/robots.txt", 11) != 0) return 0;
+	if (strstr(agent, "MSIECrawler")) return 0;
+	return !vi_counter_incr(&vih->robots, agent);
+}
+
 /* Process referer -> request pairs for web trails */
 int vi_process_web_trails(struct vih *vih, char *ref, char *req)
 {
-	int res, plen;
+	int res, plen, google;
+	char buf[VI_LINE_MAX];
+	char *src;
 
 	if (vi_is_image(req)) return 0;
-	if ((plen = vi_is_internal_link(ref))) {
-		char buf[VI_LINE_MAX];
-		char *src;
-
+	plen = vi_is_internal_link(ref);
+	google = vi_is_google_link(ref);
+	if (plen) {
 		src = (ref[plen] == '\0') ? "/" : ref+plen;
-		if (!strcmp(src, req)) return 0; /* avoid self references */
-
-		snprintf(buf, VI_LINE_MAX, "%s -> %s", src, req);
-		buf[VI_LINE_MAX-1] = '\0';
-		res = vi_counter_incr(&vih->trails, buf);
-		if (res == 0) return 1;
+	} else if (google) {
+		if (Config_graphviz_ignorenode_google) return 0;
+		src = "Google";
+	} else if (ref[0] != '\0') {
+		if (Config_graphviz_ignorenode_external) return 0;
+		src = "External Link";
+	} else {
+		if (Config_graphviz_ignorenode_noreferer) return 0;
+		src = "No Referer";
 	}
+	if (!strcmp(src, req)) return 0; /* avoid self references */
+
+	snprintf(buf, VI_LINE_MAX, "%s -> %s", src, req);
+	buf[VI_LINE_MAX-1] = '\0';
+	res = vi_counter_incr(&vih->trails, buf);
+	if (res == 0) return 1;
 	return 0;
 }
 
@@ -1161,33 +1268,46 @@ int vi_process_line(struct vih *vih, char *l)
 		vi_strlcpy(origline, l, VI_LINE_MAX);
 	/* Split the line and run all the selected processing. */
 	if (vi_parse_line(&ll, l) == 0) {
-		vi_process_date_and_hour(vih, (ll.tm.tm_wday+6)%7, ll.tm.tm_hour);
-		if (vi_process_visitors_per_day(vih, ll.host, ll.agent, ll.date, ll.ref, ll.req))
+		int seen;
+
+		if (vi_process_visitors_per_day(vih, ll.host, ll.agent,
+					ll.date, ll.ref, ll.req, &seen))
 			goto oom;
-		if (vi_process_referer(vih, ll.ref, ll.time)) goto oom;
+
+		/* The following are processed for every log line */
 		if (vi_process_page_request(vih, ll.req)) goto oom;
 		if (Config_process_error404 &&
 		    vi_process_error404(vih, origline, ll.req)) goto oom;
+		if (Config_process_google &&
+		    vi_process_googled(vih, ll.req, ll.agent, ll.time))
+			goto oom;
+		if (Config_process_web_trails &&
+		    vi_process_web_trails(vih, ll.ref, ll.req)) goto oom;
+		if (Config_process_google_keyphrases &&
+		    vi_process_google_keyphrases(vih, ll.ref, ll.time))
+			goto oom;
+
+		/* The following are processed only for new visits */
+		if (seen) return 0;
+		vi_process_date_and_hour(vih, (ll.tm.tm_wday+6)%7,
+				ll.tm.tm_hour);
+		vi_process_month_and_day(vih, ll.tm.tm_mon, ll.tm.tm_mday-1);
+		if (vi_process_referer(vih, ll.ref, ll.time)) goto oom;
 		if (Config_process_agents &&
 		    vi_process_agents(vih, ll.agent)) goto oom;
 		if (Config_process_os &&
 		    vi_process_os(vih, ll.agent)) goto oom;
 		if (Config_process_browsers &&
 		    vi_process_browsers(vih, ll.agent)) goto oom;
-		if (Config_process_google &&
-		    vi_process_googled(vih, ll.req, ll.agent, ll.time))
-			goto oom;
-		if (Config_process_google_keyphrases &&
-		    vi_process_google_keyphrases(vih, ll.ref)) goto oom;
-		if (Config_process_web_trails &&
-		    vi_process_web_trails(vih, ll.ref, ll.req)) goto oom;
 		if (Config_process_tld &&
 		    vi_process_tld(vih, ll.host)) goto oom;
+		if (Config_process_robots &&
+		    vi_process_robots(vih, ll.req, ll.agent)) goto oom;
+		return 0;
 	} else {
 		vih->invalid++;
 		return 0;
 	}
-	return 0;
 oom:
 	vi_set_error(vih, "Out of memory processing data");
 	return 1;
@@ -1574,7 +1694,7 @@ void om_html_print_header(FILE *fp)
 "	border-bottom: 1px dashed #bcbcbc;\n"
 "}\n"
 ".keyentrywe {\n"
-"	background-color: #f0f0f0;\n"
+"	background-color: #f0f090;\n"
 "	font-size: 10px;\n"
 "	padding-left: 2px;\n"
 "	border-bottom: 1px dashed #bcbcbc;\n"
@@ -1859,6 +1979,39 @@ int qsort_cmp_dates_key(const void *a, const void *b)
 	return qsort_cmp_dates_generic(a, b, 0, 1);
 }
 
+/* Compare dates (only the month/year part) in the log format:
+ * hashtable key part version */
+int qsort_cmp_months_key(const void *a, const void *b)
+{
+	int ret;
+	char dateA[VI_DATE_MAX];
+	char dateB[VI_DATE_MAX];
+	void *savedA, *savedB; /* backups of the original pointers */
+	void **A = (void**) a;
+	void **B = (void**) b;
+
+	/* We use an hack here, in order to call qsort_cmp_dates_generic
+	 * even in this case, we substitute the hashtable entries
+	 * with versions of the strings prefixed with "01", so they
+	 * will be parseble by parse_date().
+	 * In pratice for "May/2004" we instead put "01/May/2004" and so on. */
+	savedA = *A;
+	savedB = *B;
+	dateA[0] = dateB[0] = '0';
+	dateA[1] = dateB[1] = '1';
+	dateA[2] = dateB[2] = '/';
+	dateA[3] = dateB[3] = '\0';
+	vi_strlcat(dateA, (char*)*A, VI_DATE_MAX);
+	vi_strlcat(dateB, (char*)*B, VI_DATE_MAX);
+	*A = dateA;
+	*B = dateB;
+	ret = qsort_cmp_dates_generic(a, b, 0, 1);
+	/* Restore */
+	*A = savedA;
+	*B = savedB;
+	return ret;
+}
+
 /* Compare dates in the log format: hashtable value part version.
  * this sorts in reverse order, more recent dates first. */
 int qsort_cmp_dates_value(const void *a, const void *b)
@@ -1891,6 +2044,7 @@ int qsort_cmp_time_value(const void *a, const void *b)
 void vi_print_visits_report(FILE *fp, struct vih *vih)
 {
 	int days = ht_used(&vih->date), i, tot = 0, max = 0;
+	int months;
 	void **table;
 
 	Output->print_title(fp, "Unique visitors in each day");
@@ -1917,12 +2071,41 @@ void vi_print_visits_report(FILE *fp, struct vih *vih)
 		Output->print_numkeybar_entry(fp, key, max, tot, value);
 	}
 	free(table);
+
+	/* Montly */
+	if (Config_process_monthly_visitors == 0) return;
+	tot = max = 0;
+	months = ht_used(&vih->month);
+	Output->print_title(fp, "Unique visitors in each month");
+	Output->print_subtitle(fp, "Multiple hits with the same IP, user agent and access day, are considered a single visit");
+	Output->print_numkey_info(fp, "Number of unique visitors",
+			ht_used(&vih->visitors));
+	Output->print_numkey_info(fp, "Different months in logfile",
+			ht_used(&vih->month));
+	
+	if ((table = ht_get_array(&vih->month)) == NULL) {
+		fprintf(stderr, "Out Of Memory in print_visits_report()\n");
+		return;
+	}
+	qsort(table, months, sizeof(void*)*2, qsort_cmp_months_key);
+	for (i = 0; i < months; i++) {
+		long value = (long) table[(i*2)+1];
+		if (value > max)
+			max = value;
+		tot += value;
+	}
+	for (i = 0; i < months; i++) {
+		char *key = table[i*2];
+		long value = (long) table[(i*2)+1];
+		Output->print_numkeybar_entry(fp, key, max, tot, value);
+	}
+	free(table);
 }
 
 /* A report to compare visits originating from google VS all the rest. */
 void vi_print_googlevisits_report(FILE *fp, struct vih *vih)
 {
-	int days = ht_used(&vih->date), i;
+	int days = ht_used(&vih->date), i, months;
 	void **table;
 
 	Output->print_title(fp, "Unique visitors from Google in each day");
@@ -1948,6 +2131,33 @@ void vi_print_googlevisits_report(FILE *fp, struct vih *vih)
 		Output->print_numkeycomparativebar_entry(fp, key, value, googlevalue);
 	}
 	free(table);
+
+	/* Montly */
+	if (Config_process_monthly_visitors == 0) return;
+	months = ht_used(&vih->month);
+	Output->print_title(fp, "Unique visitors from Google in each month");
+	Output->print_subtitle(fp, "The red part of the bar expresses the percentage of visits originated from Google");
+	Output->print_numkey_info(fp, "Number of unique visitors",
+			ht_used(&vih->visitors));
+	Output->print_numkey_info(fp, "Number of unique visitors from google",
+			ht_used(&vih->googlevisitors));
+	Output->print_numkey_info(fp, "Different months in logfile",
+			ht_used(&vih->month));
+	
+	if ((table = ht_get_array(&vih->month)) == NULL) {
+		fprintf(stderr, "Out Of Memory in print_visits_report()\n");
+		return;
+	}
+	qsort(table, months, sizeof(void*)*2, qsort_cmp_months_key);
+	for (i = 0; i < months; i++) {
+		char *key = table[i*2];
+		long value = (long) table[(i*2)+1];
+		long googlevalue;
+
+		googlevalue = vi_counter_val(&vih->googlemonth, key);
+		Output->print_numkeycomparativebar_entry(fp, key, value, googlevalue);
+	}
+	free(table);
 }
 
 void vi_print_generic_keyval_report(FILE *fp, char *title, char *subtitle,
@@ -2063,7 +2273,7 @@ void vi_print_referers_report(FILE *fp, struct vih *vih)
 	vi_print_generic_keyval_report(
 			fp,
 			"Referers",
-			"Referers ordered by hits (referers from google excluded)",
+			"Referers ordered by visits (referers from google excluded)",
 			"Different referers",
 			Config_max_referers,
 			&vih->referers,
@@ -2123,7 +2333,7 @@ void vi_print_agents_report(FILE *fp, struct vih *vih)
 	vi_print_generic_keyval_report(
 			fp,
 			"User agents",
-			"The entire user agent string ordered by hits",
+			"The entire user agent string ordered by visits",
 			"Different agents",
 			Config_max_agents,
 			&vih->agents,
@@ -2135,7 +2345,7 @@ void vi_print_os_report(FILE *fp, struct vih *vih)
 	vi_print_generic_keyvalbar_report(
 			fp,
 			"Operating Systems",
-			"Operating Systems by hits",
+			"Operating Systems by visits",
 			"Different operating systems listed",
 			100,
 			&vih->os,
@@ -2147,7 +2357,7 @@ void vi_print_browsers_report(FILE *fp, struct vih *vih)
 	vi_print_generic_keyvalbar_report(
 			fp,
 			"Browsers",
-			"Browsers used by hits",
+			"Browsers used by visits",
 			"Different browsers listed",
 			100,
 			&vih->browsers,
@@ -2171,7 +2381,7 @@ void vi_print_google_keyphrases_report(FILE *fp, struct vih *vih)
 	vi_print_keyphrases_report(
 			fp,
 			"Google Keyphrases",
-			"Keyphrases used in google searches ordered by hits",
+			"Keyphrases used in google searches ordered by visits",
 			"Total number of keyphrases",
 			Config_max_google_keyphrases,
 			&vih->googlekeyphrases,
@@ -2183,13 +2393,25 @@ void vi_print_tld_report(FILE *fp, struct vih *vih)
 	vi_print_generic_keyvalbar_report(
 			fp,
 			"Domains",
-			"Top Level Domains sorted by hits",
+			"Top Level Domains sorted by visits",
 			"Total number of Top Level Domains",
 			Config_max_tld,
 			&vih->tld,
 			qsort_cmp_long_value);
 }
 
+void vi_print_robots_report(FILE *fp, struct vih *vih)
+{
+	vi_print_generic_keyval_report(
+			fp,
+			"Robots and web spiders",
+			"Agents requesting robots.txt. MSIECrawler excluded.",
+			"Total number of different robots",
+			Config_max_robots,
+			&vih->robots,
+			qsort_cmp_long_value);
+}
+
 /* Print a generic report where the two report items are strings
  * (usually url and date). Used to print the 'googled' and 'referers age'
  * reports. */
@@ -2242,14 +2464,26 @@ void vi_print_referers_age_report(FILE *fp, struct vih *vih)
 {
 	vi_print_generic_keytime_report(
 			fp,
-			"Referers by age",
-			"Referers ordered by age, newer on top (referers from google excluded)",
+			"Referers by first time",
+			"Referers ordered by first time date, newer on top (referers from google excluded)",
 			"Different referers",
 			Config_max_referers_age,
 			&vih->referersage,
 			qsort_cmp_time_value);
 }
 
+void vi_print_google_keyphrases_age_report(FILE *fp, struct vih *vih)
+{
+	vi_print_generic_keytime_report(
+			fp,
+			"Google Keyphrases by first time",
+			"Keyphrases ordered by first time date, newer on top",
+			"Different referers",
+			Config_max_google_keyphrases_age,
+			&vih->googlekeyphrasesage,
+			qsort_cmp_time_value);
+}
+
 void vi_print_information_report(FILE *fp, struct vih *vih)
 {
 	char buf[VI_LINE_MAX];
@@ -2267,13 +2501,17 @@ void vi_print_report_links(FILE *fp)
 {
 	void *l[] = {
 	"Unique visitors in each day", NULL,
+	"Unique visitors in each month", &Config_process_monthly_visitors,
 	"Unique visitors from Google in each day", NULL,
+	"Unique visitors from Google in each month", &Config_process_monthly_visitors,
 	"Pageviews per visit", &Config_process_pageviews,
 	"Weekday-Hour combined map", &Config_process_weekdayhour_map,
+	"Month-Day combined map", &Config_process_monthday_map,
 	"Requested pages", NULL,
 	"Requested images and CSS", NULL,
 	"Referers", NULL,
-	"Referers by age", &Config_process_referers_age,
+	"Referers by first time", &Config_process_referers_age,
+	"Robots and web spiders", &Config_process_robots,
 	"User agents", &Config_process_agents,
 	"Operating Systems", &Config_process_os,
 	"Browsers", &Config_process_browsers,
@@ -2281,6 +2519,7 @@ void vi_print_report_links(FILE *fp)
 	"Domains", &Config_process_tld,
 	"Googled pages", &Config_process_google,
 	"Google Keyphrases", &Config_process_google_keyphrases,
+	"Google Keyphrases by first time", &Config_process_google_keyphrases_age,
 	"Web trails", &Config_process_web_trails,
 	"Weekday distribution", NULL,
 	"Hours distribution", NULL,
@@ -2332,6 +2571,41 @@ void vi_print_weekdayhour_map_report(FILE *fp, struct vih *vih)
 	Output->print_bidimentional_map(fp, 24, 7, xlabel, ylabel, hw);
 }
 
+void vi_print_monthday_map_report(FILE *fp, struct vih *vih)
+{
+	char *xlabel[31] = {
+		"01", "02", "03", "04", "05", "06", "07", "08",
+		"09", "10", "11", "12", "13", "14", "15", "16",
+		"17", "18", "19", "20", "21", "22", "23", "24",
+		"25", "26", "27", "28", "29", "30", "31"};
+	char *ylabel[12] = {
+		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
+		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
+	};
+	int j, minj = 0, maxj = 0;
+	int *md = (int*) vih->monthday;
+	char buf[VI_LINE_MAX];
+
+	/* Check idexes of minimum and maximum in the array. */
+	for (j = 0; j < 12*31; j++) {
+		if (md[j] > md[maxj])
+			maxj = j;
+		if (md[j] != 0 && (md[j] < md[minj] || md[minj] == 0))
+			minj = j;
+	}
+
+	Output->print_title(fp, "Month-Day combined map");
+	Output->print_subtitle(fp, "Brighter means higher level of hits");
+	snprintf(buf, VI_LINE_MAX, "Day with max traffic is %s %s with hits",
+			ylabel[maxj/31], xlabel[maxj%31]);
+	Output->print_numkey_info(fp, buf, md[maxj]);
+	snprintf(buf, VI_LINE_MAX, "Day with min traffic is %s %s with hits",
+			ylabel[minj/31], xlabel[minj%31]);
+	Output->print_numkey_info(fp, buf, md[minj]);
+	Output->print_hline(fp);
+	Output->print_bidimentional_map(fp, 31, 12, xlabel, ylabel, md);
+}
+
 void vi_print_hline(FILE *fp)
 {
 	Output->print_hline(fp);
@@ -2390,6 +2664,10 @@ int vi_print_report(char *of, struct vih *vih)
 		vi_print_weekdayhour_map_report(fp, vih);
 		vi_print_hline(fp);
 	}
+	if (Config_process_monthday_map) {
+		vi_print_monthday_map_report(fp, vih);
+		vi_print_hline(fp);
+	}
 	if (Config_process_pageviews) {
 		vi_print_pageviews_report(fp, vih);
 		vi_print_hline(fp);
@@ -2404,6 +2682,10 @@ int vi_print_report(char *of, struct vih *vih)
 		vi_print_referers_age_report(fp, vih);
 		vi_print_hline(fp);
 	}
+	if (Config_process_robots) {
+		vi_print_robots_report(fp, vih);
+		vi_print_hline(fp);
+	}
 	if (Config_process_agents) {
 		vi_print_agents_report(fp, vih);
 		vi_print_hline(fp);
@@ -2432,6 +2714,10 @@ int vi_print_report(char *of, struct vih *vih)
 		vi_print_google_keyphrases_report(fp, vih);
 		vi_print_hline(fp);
 	}
+	if (Config_process_google_keyphrases) {
+		vi_print_google_keyphrases_age_report(fp, vih);
+		vi_print_hline(fp);
+	}
 	if (Config_process_web_trails) {
 		vi_print_trails_report(fp, vih);
 		vi_print_hline(fp);
@@ -2451,7 +2737,7 @@ int vi_print_report(char *of, struct vih *vih)
 /* ------------------------- graphviz graph generation ---------------------- */
 void vi_print_graphviz(struct vih *vih)
 {
-	int items = ht_used(&vih->trails), i, max = 0;
+	int items = ht_used(&vih->trails), i, max = 0, tot = 0;
 	void **table;
 
 	printf("digraph webtrails {\n");
@@ -2465,22 +2751,32 @@ void vi_print_graphviz(struct vih *vih)
 	qsort(table, items, sizeof(void*)*2, qsort_cmp_long_value);
 	for (i = 0; i < items; i++) {
 		long value = (long) table[(i*2)+1];
-		if (i > Config_max_trails) break;
+		tot += value;
+		if (i > Config_max_trails) continue;
 		if (max < value)
 			max = value;
 	}
+	if (max == 0) max = 1; /* avoid division by zero */
+	if (tot == 0) tot = 1;
 	for (i = 0; i < items; i++) {
 		int color;
 		char *key = table[i*2];
 		char *t;
 		long value = (long) table[(i*2)+1];
+		float percentage = ((float)value/tot)*100;
 		if (i > Config_max_trails) break;
 		color = (value*255)/max;
 		t = strstr(key, " -> ");
 		*t = '\0'; /* alter */
-		printf("\t\"%s\" -> \"%s\" [color=\"#%02X00%02X\"]\n", key, t+4, color, 255-color);
+		printf("\t\"%s\" -> \"%s\" [color=\"#%02X00%02X\" label=\"%.2f\"]\n", key, t+4, color, 255-color, percentage);
 		*t = ' '; /* restore */
 	}
+	if (!Config_graphviz_ignorenode_google)
+		printf("\tGoogle [color=\"#c0ffc0\"]\n");
+	if (!Config_graphviz_ignorenode_external)
+		printf("\t\"External Link\" [color=\"#c0ffc0\"]\n");
+	if (!Config_graphviz_ignorenode_noreferer)
+		printf("\t\"No Referer\" [color=\"#c0ffc0\"]\n");
 	free(table);
 	printf("}\n");
 }
@@ -2522,7 +2818,7 @@ void vi_stream_mode(struct vih *vih)
 /* ----------------------------------- main --------------------------------- */
 
 /* command line switche IDs */
-enum { OPT_MAXREFERERS, OPT_MAXPAGES, OPT_MAXIMAGES, OPT_USERAGENTS, OPT_ALL, OPT_MAXLINES, OPT_GOOGLE, OPT_MAXGOOGLED, OPT_MAXUSERAGENTS, OPT_OUTPUT, OPT_VERSION, OPT_HELP, OPT_PREFIX, OPT_TRAILS, OPT_GOOGLEKEYPHRASES, OPT_MAXGOOGLEKEYPHRASES, OPT_MAXTRAILS, OPT_GRAPHVIZ, OPT_WEEKDAYHOUR_MAP, OPT_REFERERSAGE, OPT_MAXREFERERSAGE, OPT_TAIL, OPT_TLD, OPT_MAXTLD, OPT_STREAM, OPT_OUTPUTFILE, OPT_UPDATEEVERY, OPT_RESETEVERY, OPT_OS, OPT_BROWSERS, OPT_ERROR404, OPT_MAXERROR404, OPT_TIMEDELTA, OPT_PAGEVIEWS};
+enum { OPT_MAXREFERERS, OPT_MAXPAGES, OPT_MAXIMAGES, OPT_USERAGENTS, OPT_ALL, OPT_MAXLINES, OPT_GOOGLE, OPT_MAXGOOGLED, OPT_MAXUSERAGENTS, OPT_OUTPUT, OPT_VERSION, OPT_HELP, OPT_PREFIX, OPT_TRAILS, OPT_GOOGLEKEYPHRASES, OPT_GOOGLEKEYPHRASESAGE, OPT_MAXGOOGLEKEYPHRASES, OPT_MAXGOOGLEKEYPHRASESAGE, OPT_MAXTRAILS, OPT_GRAPHVIZ, OPT_WEEKDAYHOUR_MAP, OPT_MONTHDAY_MAP, OPT_REFERERSAGE, OPT_MAXREFERERSAGE, OPT_TAIL, OPT_TLD, OPT_MAXTLD, OPT_STREAM, OPT_OUTPUTFILE, OPT_UPDATEEVERY, OPT_RESETEVERY, OPT_OS, OPT_BROWSERS, OPT_ERROR404, OPT_MAXERROR404, OPT_TIMEDELTA, OPT_PAGEVIEWS, OPT_ROBOTS, OPT_MAXROBOTS, OPT_GRAPHVIZ_ignorenode_GOOGLE, OPT_GRAPHVIZ_ignorenode_EXTERNAL, OPT_GRAPHVIZ_ignorenode_NOREFERER};
 
 /* command line switches definition:
  * the rule with short options is to take upper case the
@@ -2532,14 +2828,17 @@ static struct ago_optlist visitors_optlist[] = {
 	{ 'T',  "trails",		OPT_TRAILS,		AGO_NOARG},
 	{ 'G',	"google",		OPT_GOOGLE,		AGO_NOARG},
 	{ 'K',	"google-keyphrases",	OPT_GOOGLEKEYPHRASES,	AGO_NOARG},
+	{ 'Z',	"google-keyphrases-age", OPT_GOOGLEKEYPHRASESAGE, AGO_NOARG},
 	{ 'U',	"user-agents",		OPT_USERAGENTS,		AGO_NOARG},
 	{ 'W',  "weekday-hour-map",	OPT_WEEKDAYHOUR_MAP,	AGO_NOARG},
+	{ 'M',  "month-day-map",	OPT_MONTHDAY_MAP,	AGO_NOARG},
 	{ 'R',  "referers-age",		OPT_REFERERSAGE,	AGO_NOARG},
 	{ 'D',  "domains",		OPT_TLD,		AGO_NOARG},
 	{ 'O',  "operating-systems",	OPT_OS,			AGO_NOARG},
 	{ 'B',  "browsers",		OPT_BROWSERS,		AGO_NOARG},
 	{ 'X',  "error404",		OPT_ERROR404,		AGO_NOARG},
-	{ 'P',  "pageviews",		OPT_PAGEVIEWS,		AGO_NOARG},
+	{ 'Y',  "pageviews",		OPT_PAGEVIEWS,		AGO_NOARG},
+	{ 'S',	"robots",		OPT_ROBOTS,		AGO_NOARG},
 	{ '\0', "stream",		OPT_STREAM,		AGO_NOARG},
 	{ '\0', "update-every",		OPT_UPDATEEVERY,	AGO_NEEDARG},
 	{ '\0',	"reset-every",		OPT_RESETEVERY,		AGO_NEEDARG},
@@ -2553,11 +2852,20 @@ static struct ago_optlist visitors_optlist[] = {
 	{ 't',	"max-trails",		OPT_MAXTRAILS,		AGO_NEEDARG},
 	{ 'g',	"max-googled",		OPT_MAXGOOGLED,		AGO_NEEDARG},
 	{ 'k',	"max-google-keyphrases",OPT_MAXGOOGLEKEYPHRASES,AGO_NEEDARG},
+	{ 'z',	"max-google-keyphrases-age",OPT_MAXGOOGLEKEYPHRASESAGE,
+		AGO_NEEDARG},
 	{ 'a',	"max-referers-age",	OPT_MAXREFERERSAGE,	AGO_NEEDARG},
 	{ 'd',	"max-domains",		OPT_MAXTLD,		AGO_NEEDARG},
+	{ 's',	"max-robots",		OPT_MAXROBOTS,		AGO_NEEDARG},
 	{ 'P',  "prefix",		OPT_PREFIX,		AGO_NEEDARG},
 	{ 'o',  "output",		OPT_OUTPUT,		AGO_NEEDARG},
 	{ 'V',  "graphviz",		OPT_GRAPHVIZ,		AGO_NOARG},
+	{ '\0', "graphviz-ignorenode-google", OPT_GRAPHVIZ_ignorenode_GOOGLE,
+		AGO_NOARG},
+	{ '\0', "graphviz-ignorenode-external", OPT_GRAPHVIZ_ignorenode_EXTERNAL,
+		AGO_NOARG},
+	{ '\0', "graphviz-ignorenode-noreferer", OPT_GRAPHVIZ_ignorenode_NOREFERER,
+		AGO_NOARG},
 	{ 'v',  "version",		OPT_VERSION,		AGO_NOARG},
 	{ '\0', "tail",			OPT_TAIL,		AGO_NOARG},
 	{ '\0', "time-delta",		OPT_TIMEDELTA,		AGO_NEEDARG},
@@ -2583,7 +2891,7 @@ void visitors_show_help(void)
 					"<argument>" : "");
 	}
 	printf("For more information visit http://www.hping.org/visitors\n"
-	       "Visitors is Copyright(C) 2004 Salvatore Sanfilippo <antirez@invece.org>\n");
+	       "Visitors is Copyright(C) 2004,2005 Salvatore Sanfilippo <antirez@invece.org>\n");
 }
 
 int main(int argc, char **argv)
@@ -2634,12 +2942,18 @@ int main(int argc, char **argv)
 		case OPT_MAXGOOGLEKEYPHRASES:
 			Config_max_google_keyphrases = atoi(ago_optarg);
 			break;
+		case OPT_MAXGOOGLEKEYPHRASESAGE:
+			Config_max_google_keyphrases_age = atoi(ago_optarg);
+			break;
 		case OPT_MAXREFERERSAGE:
 			Config_max_referers_age = atoi(ago_optarg);
 			break;
 		case OPT_MAXTLD:
 			Config_max_tld = atoi(ago_optarg);
 			break;
+		case OPT_MAXROBOTS:
+			Config_max_robots = atoi(ago_optarg);
+			break;
 		case OPT_USERAGENTS:
 			Config_process_agents = 1;
 			break;
@@ -2649,6 +2963,9 @@ int main(int argc, char **argv)
 		case OPT_GOOGLEKEYPHRASES:
 			Config_process_google_keyphrases = 1;
 			break;
+		case OPT_GOOGLEKEYPHRASESAGE:
+			Config_process_google_keyphrases_age = 1;
+			break;
 		case OPT_TLD:
 			Config_process_tld = 1;
 			break;
@@ -2664,17 +2981,23 @@ int main(int argc, char **argv)
 		case OPT_PAGEVIEWS:
 			Config_process_pageviews = 1;
 			break;
+		case OPT_ROBOTS:
+			Config_process_robots = 1;
+			break;
 		case OPT_ALL:
 			Config_process_agents = 1;
 			Config_process_google = 1;
 			Config_process_google_keyphrases = 1;
+			Config_process_google_keyphrases_age = 1;
 			Config_process_weekdayhour_map = 1;
+			Config_process_monthday_map = 1;
 			Config_process_referers_age = 1;
 			Config_process_tld = 1;
 			Config_process_os = 1;
 			Config_process_browsers = 1;
 			Config_process_error404 = 1;
 			Config_process_pageviews = 1;
+			Config_process_robots = 1;
 			break;
 		case OPT_PREFIX:
 			if (Config_prefix_num < VI_PREFIXES_MAX) {
@@ -2700,8 +3023,10 @@ int main(int argc, char **argv)
 				Config_max_googled = aux;
 				Config_max_trails = aux;
 				Config_max_google_keyphrases = aux;
+				Config_max_google_keyphrases_age = aux;
 				Config_max_referers_age = aux;
 				Config_max_tld = aux;
+				Config_max_robots = aux;
 			}
 			break;
 		case OPT_OUTPUT:
@@ -2719,12 +3044,24 @@ int main(int argc, char **argv)
 			Config_graphviz_mode = 1;
 			Config_process_web_trails = 1;
 			break;
+		case OPT_GRAPHVIZ_ignorenode_GOOGLE:
+			Config_graphviz_ignorenode_google = 1;
+			break;
+		case OPT_GRAPHVIZ_ignorenode_EXTERNAL:
+			Config_graphviz_ignorenode_external= 1;
+			break;
+		case OPT_GRAPHVIZ_ignorenode_NOREFERER:
+			Config_graphviz_ignorenode_noreferer = 1;
+			break;
 		case OPT_TAIL:
 			Config_tail_mode = 1;
 			break;
 		case OPT_WEEKDAYHOUR_MAP:
 			Config_process_weekdayhour_map = 1;
 			break;
+		case OPT_MONTHDAY_MAP:
+			Config_process_monthday_map = 1;
+			break;
 		case OPT_REFERERSAGE:
 			Config_process_referers_age = 1;
 			break;
@@ -2764,7 +3101,7 @@ int main(int argc, char **argv)
 	/* If the prefix was not set, but the user asks for
 	 * web trails, notify it and exit. */
 	if (Config_process_web_trails && !Config_prefix_num) {
-		fprintf(stderr, "At least one prefix must be specified (using --prefix) for web trails\n");
+		fprintf(stderr, "At least one prefix must be specified (using --prefix) for web trails\nExample: --prefix http://your.site.org\n");
 		exit(1);
 	}
 	/* If stream-mode is enabled, --output-file should be specified. */
@@ -2774,7 +3111,7 @@ int main(int argc, char **argv)
 	}
 	/* Set the default output module */
 	if (Output == NULL)
-		Output = &OutputModuleText;
+		Output = &OutputModuleHtml;
 	/* Change to "C" locale for date/time related functions */
 	setlocale(LC_ALL, "C");
 	/* Process all the log files specified. */
@@ -2797,5 +3134,11 @@ int main(int argc, char **argv)
 		}
 	}
 	vi_print_statistics(vih);
+        /* The following is commented in releases as to free the hashtable
+         * memory is very slow, it's better to just exit the program.
+         * Still it is important to be able to re-enable a good cleanup
+         * in order to run visitors against valgrind to check for memory
+         * leaks. */
+        /* vi_free(vih); */
 	return 0;
 }
